/** \file POperators_ZOT.h */

#ifndef P_OPERATORS_ZOT_H
#define P_OPERATORS_ZOT_H

#include "AMDiS.h"

using namespace std;
using namespace AMDiS;

/// < factor*phase*u , psi >
class Phase_ZOT : public ZeroOrderTerm 
{ 
public: 
Phase_ZOT(DOFVectorBase<double>* phaseDV, double fac=1.0);

/// Implementation of \ref OperatorTerm::initElement().
void initElement(const ElInfo* elInfo, 
		SubAssembler* subAssembler,
		Quadrature *quad=NULL);
void initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
		SubAssembler* subAssembler,
		Quadrature *quad=NULL);

inline void getC(const ElInfo *, int nPoints, ElementVector &C);

inline void eval(int nPoints,
		const mtl::dense_vector<double>&,
		const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
		const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
		mtl::dense_vector<double>& result,
		double opFactor);

virtual double f(const int iq) const;
void setFactor(const double fac_) { fac=fac_; }

protected:
DOFVectorBase<double> *phaseDV;
mtl::dense_vector<double> phase;
double fac;
}; 

/* -------------------------------------------------------------- */

/// < factor*phase*v*u , psi >
class FactorPhase_ZOT : public Phase_ZOT
{
public:
	FactorPhase_ZOT(DOFVectorBase<double> *phaseDV_, DOFVectorBase<double> *rhoDV_, double fac_=1.0);
	void initElement(const ElInfo* elInfo,
		SubAssembler* subAssembler,
		Quadrature *quad = NULL);
	void initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
		SubAssembler*,
		Quadrature *quad = NULL);
	inline double f(const int iq) const;
protected:
	DOFVectorBase<double> *rhoDV;
	mtl::dense_vector<double> rho;
};

/* -------------------------------------------------------------- */

/// < factor*(1-phase)*u , psi > or < factor*v[i]*(1-phase)*u , psi > where v in R^n
class PhaseInverse_ZOT : public ZeroOrderTerm 
{ 
public: 
PhaseInverse_ZOT(DOFVectorBase<double>* phaseDV, double fac=1.0);
PhaseInverse_ZOT(DOFVectorBase<double>* phaseDV, double fac, WorldVector<double>* facVec, int component);

/// Implementation of \ref OperatorTerm::initElement().
void initElement(const ElInfo* elInfo, 
		SubAssembler* subAssembler,
		Quadrature *quad=NULL);
void initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
		SubAssembler* subAssembler,
		Quadrature *quad=NULL);

inline void getC(const ElInfo *, int nPoints, ElementVector &C);

inline void eval(int nPoints,
		const mtl::dense_vector<double>&,
		const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
		const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
		mtl::dense_vector<double>& result,
		double opFactor);

inline double f(const int iq) const;
void setFactor(const double fac_) { fac=fac_; }

protected:
DOFVectorBase<double> *phaseDV;
mtl::dense_vector<double> phase;
double fac;
int component;
WorldVector<double>* facVec;
}; 

/* -------------------------------------------------------------- */

/// < factor*(v^3)*u , psi >
class Pow3_ZOT : public ZeroOrderTerm
{
public:
	Pow3_ZOT(DOFVectorBase<double> *rhoDV_, double fac_=1.0);
	void initElement(const ElInfo* elInfo,
		SubAssembler* subAssembler,
		Quadrature *quad = NULL);
	void initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
		SubAssembler*,
		Quadrature *quad = NULL);
inline void getC(const ElInfo *, int nPoints, ElementVector &C);

inline void eval(int nPoints,
		const mtl::dense_vector<double>&,
		const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
		const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
		mtl::dense_vector<double>& result,
		double opFactor);
	inline double f(const int iq) const;
protected:
	DOFVectorBase<double> *rhoDV;
	mtl::dense_vector<double> rho;
	double fac;
};

/* -------------------------------------------------------------- */

/// < factor*(v^3)*phi*u , psi >
class Pow3Phase_ZOT : public Phase_ZOT
{
public:
	Pow3Phase_ZOT(DOFVectorBase<double>* phaseDV, DOFVectorBase<double> *rhoDV_, double fac_=1.0);
	void initElement(const ElInfo* elInfo,
		SubAssembler* subAssembler,
		Quadrature *quad = NULL);
	void initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
		SubAssembler*,
		Quadrature *quad = NULL);
	inline double f(const int iq) const;
protected:
	DOFVectorBase<double> *rhoDV;
	mtl::dense_vector<double> rho;
};
/* -------------------------------------------------------------- */

/// < factor*(v^2)*u , psi >
class Pow2_ZOT : public ZeroOrderTerm
{
public:
	Pow2_ZOT(DOFVectorBase<double> *rhoDV_, double fac_=1.0);
	void initElement(const ElInfo* elInfo,
		SubAssembler* subAssembler,
		Quadrature *quad = NULL);
	void initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
		SubAssembler*,
		Quadrature *quad = NULL);
inline void getC(const ElInfo *, int nPoints, ElementVector &C);

inline void eval(int nPoints,
		const mtl::dense_vector<double>&,
		const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
		const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
		mtl::dense_vector<double>& result,
		double opFactor);
	inline double f(const int iq) const;
protected:
	DOFVectorBase<double> *rhoDV;
	mtl::dense_vector<double> rho;
	double fac;
};

/* -------------------------------------------------------------- */

/// < factor*(v^2)*phi*u , psi >
class Pow2Phase_ZOT : public Phase_ZOT
{
public:
	Pow2Phase_ZOT(DOFVectorBase<double>* phaseDV, DOFVectorBase<double> *rhoDV_, double fac_=1.0);
	void initElement(const ElInfo* elInfo,
		SubAssembler* subAssembler,
		Quadrature *quad = NULL);
	void initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
		SubAssembler*,
		Quadrature *quad = NULL);
	inline double f(const int iq) const;
protected:
	DOFVectorBase<double> *rhoDV;
	mtl::dense_vector<double> rho;
};

/* -------------------------------------------------------------- */

/// < factor*phase*(-1/max(1e-6, sqr(v+0.9)))*u , psi >
class ConstrainedFracSqr_ZOT : public Phase_ZOT
{
public:
	ConstrainedFracSqr_ZOT(DOFVectorBase<double> *phaseDV_, DOFVectorBase<double> *rhoDV_, double fac_=1.0);
	void initElement(const ElInfo* elInfo,
		SubAssembler* subAssembler,
		Quadrature *quad = NULL);
	void initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
		SubAssembler*,
		Quadrature *quad = NULL);
	inline double f(const int iq) const;
protected:
	DOFVectorBase<double> *rhoDV;
	mtl::dense_vector<double> rho;
};

/// < factor*phase*(1/max(1e-6, v+0.9))*u , psi >
class ConstrainedFrac_ZOT : public Phase_ZOT
{
public:
	ConstrainedFrac_ZOT(DOFVectorBase<double> *phaseDV_, DOFVectorBase<double> *rhoDV_, double fac_=1.0);
	void initElement(const ElInfo* elInfo,
		SubAssembler* subAssembler,
		Quadrature *quad = NULL);
	void initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
		SubAssembler*,
		Quadrature *quad = NULL);
	inline double f(const int iq) const;
protected:
	DOFVectorBase<double> *rhoDV;
	mtl::dense_vector<double> rho;
};

/// < (...)*u , psi >
class ConstrainedPowImpl_ZOT : public ZeroOrderTerm
{
public:
    ConstrainedPowImpl_ZOT(DOFVectorBase<double> *rhoDV_, double p_=20.0, double rhoMin_=-0.9, double fac_=1.0);
    void initElement(const ElInfo* elInfo,
        SubAssembler* subAssembler,
        Quadrature *quad = NULL);
    void initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
        SubAssembler*,
        Quadrature *quad = NULL);
    void getC(const ElInfo *, int nPoints, ElementVector &C);
    void eval(int nPoints,
            const mtl::dense_vector<double>&,
            const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
            const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
            mtl::dense_vector<double>& result,
            double opFactor);
    inline double f(const int iq) const;
protected:
    DOFVectorBase<double> *rhoDV;
    mtl::dense_vector<double> rho;
    double p,rhoMin,fac;
};

/// < (...)*u , psi >
class ConstrainedPowExpl_ZOT : public ZeroOrderTerm
{
public:
    ConstrainedPowExpl_ZOT(DOFVectorBase<double> *rhoDV_, double p_=20.0, double rhoMin_=-0.9, double fac_=1.0);
    void initElement(const ElInfo* elInfo,
        SubAssembler* subAssembler,
        Quadrature *quad = NULL);
    void initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
        SubAssembler*,
        Quadrature *quad = NULL);
    void getC(const ElInfo *, int nPoints, ElementVector &C);
    void eval(int nPoints,
            const mtl::dense_vector<double>&,
            const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
            const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
            mtl::dense_vector<double>& result,
            double opFactor);
    inline double f(const int iq) const;
protected:
    DOFVectorBase<double> *rhoDV;
    mtl::dense_vector<double> rho;
    double p,rhoMin,fac;
};

/// < (...)*u , psi >
class ConstrainedPow_ZOT : public ZeroOrderTerm
{
public:
    ConstrainedPow_ZOT(DOFVectorBase<double> *rhoDV_, double p_=20.0, double rhoMin_=-0.9, double fac_=1.0);
    void initElement(const ElInfo* elInfo,
        SubAssembler* subAssembler,
        Quadrature *quad = NULL);
    void initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
        SubAssembler*,
        Quadrature *quad = NULL);
    void getC(const ElInfo *, int nPoints, ElementVector &C);
    void eval(int nPoints,
            const mtl::dense_vector<double>&,
            const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
            const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
            mtl::dense_vector<double>& result,
            double opFactor);
    inline double f(const int iq) const;
protected:
    DOFVectorBase<double> *rhoDV;
    mtl::dense_vector<double> rho;
    double p,rhoMin,fac;
};

/* -------------------------------------------------------------- */

/// < factor*d_i(v)*u , psi > or < factor*b[j]*d_i(v)*u , psi >
class PartialDerivative_ZOT : public ZeroOrderTerm
{
public:
PartialDerivative_ZOT(DOFVectorBase<double>* vecDV_, int component_, double fac_=1.0);
PartialDerivative_ZOT(DOFVectorBase<double>* vecDV_, int component_, double fac_, WorldVector<double>* vec, int component2_);

void initElement(const ElInfo*, 
	SubAssembler*,
	Quadrature *quad= NULL);
void initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
	SubAssembler*,
	Quadrature *quad = NULL);

void getC(const ElInfo *, int nPoints, ElementVector &C);

void eval(int nPoints,
		const mtl::dense_vector<double>&,
		const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
		const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
		mtl::dense_vector<double>& result,
		double opFactor);

protected:
DOFVectorBase<double> *vecDV;
mtl::dense_vector<WorldVector<double> > grad;
double fac;
WorldVector<double>* facVec;
int component, component2;
};

/* -------------------------------------------------------------- */

/// < factor*v*d_i(w)*u , psi >
class VecAndPartialDerivative_ZOT : public ZeroOrderTerm
{
public:
VecAndPartialDerivative_ZOT(DOFVectorBase<double>* vecDV_, DOFVectorBase<double>* gradDV_, int component_, double fac_=1.0);

void initElement(const ElInfo*, 
	SubAssembler*,
	Quadrature *quad= NULL);
void initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
	SubAssembler*,
	Quadrature *quad = NULL);

void getC(const ElInfo *, int nPoints, ElementVector &C);

void eval(int nPoints,
		const mtl::dense_vector<double>&,
		const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
		const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
		mtl::dense_vector<double>& result,
		double opFactor);

protected:
DOFVectorBase<double> *vecDV, *gradDV;
mtl::dense_vector<double> vec;
double fac;
mtl::dense_vector<WorldVector<double> > grad;
int component;
};

/* -------------------------------------------------------------- */

/// < factor*v1*v2*d_i(w)*u , psi >
class Vec2AndPartialDerivative_ZOT : public ZeroOrderTerm
{
public:
Vec2AndPartialDerivative_ZOT(DOFVectorBase<double>* vec1DV_, DOFVectorBase<double>* vec2DV_, DOFVectorBase<double>* gradDV_, int component_, double fac_=1.0);

void initElement(const ElInfo*,
	SubAssembler*,
	Quadrature *quad= NULL);
void initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
	SubAssembler*,
	Quadrature *quad = NULL);

void getC(const ElInfo *, int nPoints, ElementVector &C);

void eval(int nPoints,
		const mtl::dense_vector<double>&,
		const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
		const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
		mtl::dense_vector<double>& result,
		double opFactor);

protected:
DOFVectorBase<double> *vec1DV, *vec2DV, *gradDV;
mtl::dense_vector<double> vec1, vec2;
double fac;
mtl::dense_vector<WorldVector<double> > grad;
int component;
};

/* -------------------------------------------------------------- */

/// < factor*v*d_i(w)*u , psi >
class WettingAngle_ZOT : public ZeroOrderTerm
{
public:
WettingAngle_ZOT(DOFVectorBase<double>* vecDV_, DOFVectorBase<double>* gradDV_, double fac_=1.0);

void initElement(const ElInfo*, 
	SubAssembler*,
	Quadrature *quad= NULL);
void initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
	SubAssembler*,
	Quadrature *quad = NULL);

void getC(const ElInfo *, int nPoints, ElementVector &C);

void eval(int nPoints,
		const mtl::dense_vector<double>&,
		const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
		const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
		mtl::dense_vector<double>& result,
		double opFactor);
inline double f(const int iq) const;

protected:
DOFVectorBase<double> *vecDV, *gradDV;
// mtl::dense_vector<double> vec;
double fac;
mtl::dense_vector<WorldVector<double> > grad;
mtl::dense_vector<WorldVector<double> > gradC;
};

/* -------------------------------------------------------------- */

/// < factor*v*d_i(w)*u , psi >
class WettingAngleExpl_ZOT : public WettingAngle_ZOT
{
public:
WettingAngleExpl_ZOT(DOFVectorBase<double>* vecDV_, DOFVectorBase<double>* gradDV_, double fac_=1.0);

inline double f(const int iq) const;
};

/* -------------------------------------------------------------- */

/// < factor*v*d_i(w)*u , psi >
class WettingAngleImpl_ZOT : public WettingAngle_ZOT
{
public:
WettingAngleImpl_ZOT(DOFVectorBase<double>* vecDV_, DOFVectorBase<double>* gradDV_, double fac_=1.0);

inline double f(const int iq) const;
};

/* -------------------------------------------------------------- */

class WettingAngleLowerOrder_ZOT : public ZeroOrderTerm
{
public:
/// Constructor.
WettingAngleLowerOrder_ZOT(
  DOFVectorBase<double> *chDV_, 
  DOFVectorBase<double>* phaseDV, 
  double factor_=1.0);

/// Implementation of \ref OperatorTerm::initElement().
void initElement(const ElInfo* elInfo, SubAssembler* subAssembler,
        Quadrature *quad = NULL);

/// Implementation of \ref OperatorTerm::initElement() for multilpe meshes.
void initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
        SubAssembler* subAssembler,
        Quadrature *quad = NULL);

void getC(const ElInfo *, int nPoints, ElementVector &C);

void eval(int nPoints,
    const mtl::dense_vector<double>&,
    const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
    const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
    mtl::dense_vector<double>& result,
    double opFactor);
inline double f(const int iq) const;

protected:

/** \brief
* DOFVector to be evaluated at quadrature points.
*/
DOFVectorBase<double>* chDV;
DOFVectorBase<double>* phaseDV;

/** \brief
* Vector v at quadrature points.
*/
mtl::dense_vector<double> ch;
mtl::dense_vector<double> phase;
mtl::dense_vector<WorldVector<double> > grdCh;
mtl::dense_vector<WorldVector<double> > grdPhase;

double fac;
};

/* -------------------------------------------------------------- */


/* -------------------------------------------------------------- */

/// < factor*v*d_i(w)*u , psi >
class WettingAngle_Sharp_ZOT : public ZeroOrderTerm
{
public:
WettingAngle_Sharp_ZOT(DOFVectorBase<double>* vecDV_, double fac_=1.0);

void initElement(const ElInfo*, 
  SubAssembler*,
  Quadrature *quad= NULL);
void initElement(const ElInfo* largeElInfo, const ElInfo* smallElInfo,
  SubAssembler*,
  Quadrature *quad = NULL);

void getC(const ElInfo *, int nPoints, ElementVector &C);

void eval(int nPoints,
    const mtl::dense_vector<double>&,
    const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
    const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
    mtl::dense_vector<double>& result,
    double opFactor);
inline double f(const int iq) const;

protected:
DOFVectorBase<double> *vecDV;
mtl::dense_vector<WorldVector<double> > grad;
double fac;
};

/* -------------------------------------------------------------- */

/// < factor*v*d_i(w)*u , psi >
class WettingAngleExpl_Sharp_ZOT : public WettingAngle_Sharp_ZOT
{
public:
WettingAngleExpl_Sharp_ZOT(DOFVectorBase<double>* vecDV_, double fac_=1.0);

inline double f(const int iq) const;
};

/* -------------------------------------------------------------- */

/// < factor*v*d_i(w)*u , psi >
class WettingAngleImpl_Sharp_ZOT : public WettingAngle_Sharp_ZOT
{
public:
WettingAngleImpl_Sharp_ZOT(DOFVectorBase<double>* vecDV_, double fac_=1.0);

inline double f(const int iq) const;
};

#endif // P_OPERATORS_ZOT_H
